@@ -6,6 +6,8 @@
  *  This file written by Ryan C. Gordon.
  */
 
+#define _CRT_SECURE_NO_WARNINGS 1
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
@@ -23,11 +25,14 @@
 
 #include <time.h>
 
+/* Define this, so the compiler doesn't complain about using old APIs. */
+#define PHYSFS_DEPRECATED
+
 #include "physfs.h"
 
-#define TEST_VERSION_MAJOR  2
+#define TEST_VERSION_MAJOR  3
 #define TEST_VERSION_MINOR  0
-#define TEST_VERSION_PATCH  3
+#define TEST_VERSION_PATCH  1
 
 static FILE *history_file = NULL;
 static PHYSFS_uint32 do_buffer_size = 0;
@@ -64,6 +69,8 @@ static void output_archivers(void)
             printf(" * %s: %s\n    Written by %s.\n    %s\n",
                     (*i)->extension, (*i)->description,
                     (*i)->author, (*i)->url);
+            printf("    %s symbolic links.\n",
+                    (*i)->supportsSymlinks ? "Supports" : "Does not support");
         } /* for */
     } /* else */
 
@@ -73,7 +80,7 @@ static void output_archivers(void)
 
 static int cmd_quit(char *args)
 {
-    return(0);
+    return 0;
 } /* cmd_quit */
 
 
@@ -90,7 +97,7 @@ static int cmd_init(char *args)
     else
         printf("Failure. reason: %s.\n", PHYSFS_getLastError());
 
-    return(1);
+    return 1;
 } /* cmd_init */
 
 
@@ -101,7 +108,7 @@ static int cmd_deinit(char *args)
     else
         printf("Failure. reason: %s.\n", PHYSFS_getLastError());
 
-    return(1);
+    return 1;
 } /* cmd_deinit */
 
 
@@ -119,20 +126,34 @@ static int cmd_addarchive(char *args)
 
     /*printf("[%s], [%d]\n", args, appending);*/
 
-    if (PHYSFS_addToSearchPath(args, appending))
+    if (PHYSFS_mount(args, NULL, appending))
         printf("Successful.\n");
     else
         printf("Failure. reason: %s.\n", PHYSFS_getLastError());
 
-    return(1);
+    return 1;
 } /* cmd_addarchive */
 
 
-static int cmd_mount(char *args)
+/* wrap free() to avoid calling convention wankery. */
+static void freeBuf(void *buf)
+{
+    free(buf);
+} /* freeBuf */
+
+typedef enum
+{
+    MNTTYPE_PATH,
+    MNTTYPE_MEMORY,
+    MNTTYPE_HANDLE
+} MountType;
+
+static int cmd_mount_internal(char *args, const MountType mnttype)
 {
     char *ptr;
     char *mntpoint = NULL;
     int appending = 0;
+    int rc = 0;
 
     if (*args == '\"')
     {
@@ -141,7 +162,7 @@ static int cmd_mount(char *args)
         if (ptr == NULL)
         {
             printf("missing string terminator in argument.\n");
-            return(1);
+            return 1;
         } /* if */
         *(ptr) = '\0';
     } /* if */
@@ -159,7 +180,7 @@ static int cmd_mount(char *args)
         if (ptr == NULL)
         {
             printf("missing string terminator in argument.\n");
-            return(1);
+            return 1;
         } /* if */
         *(ptr) = '\0';
     } /* if */
@@ -172,15 +193,101 @@ static int cmd_mount(char *args)
 
     /*printf("[%s], [%s], [%d]\n", args, mntpoint, appending);*/
 
-    if (PHYSFS_mount(args, mntpoint, appending))
+    if (mnttype == MNTTYPE_PATH)
+        rc = PHYSFS_mount(args, mntpoint, appending);
+
+    else if (mnttype == MNTTYPE_HANDLE)
+    {
+        PHYSFS_File *f = PHYSFS_openRead(args);
+        if (f == NULL)
+        {
+            printf("PHYSFS_openRead('%s') failed. reason: %s.\n", args, PHYSFS_getLastError());
+            return 1;
+        } /* if */
+
+        rc = PHYSFS_mountHandle(f, args, mntpoint, appending);
+        if (!rc)
+            PHYSFS_close(f);
+    } /* else if */
+
+    else if (mnttype == MNTTYPE_MEMORY)
+    {
+        FILE *in = fopen(args, "rb");
+        void *buf = NULL;
+        long len = 0;
+
+        if (in == NULL)
+        {
+            printf("Failed to open %s to read into memory: %s.\n", args, strerror(errno));
+            return 1;
+        } /* if */
+
+        if ( (fseek(in, 0, SEEK_END) != 0) || ((len = ftell(in)) < 0) )
+        {
+            printf("Failed to find size of %s to read into memory: %s.\n", args, strerror(errno));
+            fclose(in);
+            return 1;
+        } /* if */
+
+        buf = malloc(len);
+        if (buf == NULL)
+        {
+            printf("Failed to allocate space to read %s into memory: %s.\n", args, strerror(errno));
+            fclose(in);
+            return 1;
+        } /* if */
+
+        if ((fseek(in, 0, SEEK_SET) != 0) || (fread(buf, len, 1, in) != 1))
+        {
+            printf("Failed to read %s into memory: %s.\n", args, strerror(errno));
+            fclose(in);
+            free(buf);
+            return 1;
+        } /* if */
+
+        fclose(in);
+
+        rc = PHYSFS_mountMemory(buf, len, freeBuf, args, mntpoint, appending);
+    } /* else */
+
+    if (rc)
         printf("Successful.\n");
     else
         printf("Failure. reason: %s.\n", PHYSFS_getLastError());
 
-    return(1);
+    return 1;
+} /* cmd_mount_internal */
+
+
+static int cmd_mount(char *args)
+{
+    return cmd_mount_internal(args, MNTTYPE_PATH);
 } /* cmd_mount */
 
 
+static int cmd_mount_mem(char *args)
+{
+    return cmd_mount_internal(args, MNTTYPE_MEMORY);
+} /* cmd_mount_mem */
+
+
+static int cmd_mount_handle(char *args)
+{
+    return cmd_mount_internal(args, MNTTYPE_HANDLE);
+} /* cmd_mount_handle */
+
+static int cmd_getmountpoint(char *args)
+{
+    if (*args == '\"')
+    {
+        args++;
+        args[strlen(args) - 1] = '\0';
+    } /* if */
+
+    printf("Dir [%s] is mounted at [%s].\n", args, PHYSFS_getMountPoint(args));
+    return 1;
+} /* cmd_getmountpoint */
+
 static int cmd_removearchive(char *args)
 {
     if (*args == '\"')
@@ -189,12 +296,12 @@ static int cmd_removearchive(char *args)
         args[strlen(args) - 1] = '\0';
     } /* if */
 
-    if (PHYSFS_removeFromSearchPath(args))
+    if (PHYSFS_unmount(args))
         printf("Successful.\n");
     else
         printf("Failure. reason: %s.\n", PHYSFS_getLastError());
 
-    return(1);
+    return 1;
 } /* cmd_removearchive */
 
 
@@ -223,21 +330,21 @@ static int cmd_enumerate(char *args)
         PHYSFS_freeList(rc);
     } /* else */
 
-    return(1);
+    return 1;
 } /* cmd_enumerate */
 
 
 static int cmd_getdirsep(char *args)
 {
     printf("Directory separator is [%s].\n", PHYSFS_getDirSeparator());
-    return(1);
+    return 1;
 } /* cmd_getdirsep */
 
 
 static int cmd_getlasterror(char *args)
 {
     printf("last error is [%s].\n", PHYSFS_getLastError());
-    return(1);
+    return 1;
 } /* cmd_getlasterror */
 
 
@@ -258,7 +365,7 @@ static int cmd_getcdromdirs(char *args)
         PHYSFS_freeList(rc);
     } /* else */
 
-    return(1);
+    return 1;
 } /* cmd_getcdromdirs */
 
 
@@ -279,28 +386,41 @@ static int cmd_getsearchpath(char *args)
         PHYSFS_freeList(rc);
     } /* else */
 
-    return(1);
+    return 1;
 } /* cmd_getcdromdirs */
 
 
 static int cmd_getbasedir(char *args)
 {
     printf("Base dir is [%s].\n", PHYSFS_getBaseDir());
-    return(1);
+    return 1;
 } /* cmd_getbasedir */
 
 
 static int cmd_getuserdir(char *args)
 {
     printf("User dir is [%s].\n", PHYSFS_getUserDir());
-    return(1);
+    return 1;
 } /* cmd_getuserdir */
 
 
+static int cmd_getprefdir(char *args)
+{
+    char *org;
+    char *appName;
+    char *ptr = args;
+
+    org = ptr;
+    ptr = strchr(ptr, ' '); *ptr = '\0'; ptr++; appName = ptr;
+    printf("Pref dir is [%s].\n", PHYSFS_getPrefDir(org, appName));
+    return 1;
+} /* cmd_getprefdir */
+
+
 static int cmd_getwritedir(char *args)
 {
     printf("Write dir is [%s].\n", PHYSFS_getWriteDir());
-    return(1);
+    return 1;
 } /* cmd_getwritedir */
 
 
@@ -317,7 +437,7 @@ static int cmd_setwritedir(char *args)
     else
         printf("Failure. reason: %s.\n", PHYSFS_getLastError());
 
-    return(1);
+    return 1;
 } /* cmd_setwritedir */
 
 
@@ -334,7 +454,7 @@ static int cmd_permitsyms(char *args)
     num = atoi(args);
     PHYSFS_permitSymbolicLinks(num);
     printf("Symlinks are now %s.\n", num ? "permitted" : "forbidden");
-    return(1);
+    return 1;
 } /* cmd_permitsyms */
 
 
@@ -358,7 +478,7 @@ static int cmd_setbuffer(char *args)
         printf("Further tests will NOT use a buffer.\n");
     } /* else */
 
-    return(1);
+    return 1;
 } /* cmd_setbuffer */
 
 
@@ -395,7 +515,7 @@ static int cmd_stressbuffer(char *args)
                 printf("PHYSFS_setBuffer() failed: %s.\n", PHYSFS_getLastError());
                 PHYSFS_close(f);
                 PHYSFS_delete("test.txt");
-                return(1);
+                return 1;
             } /* if */
 
             strcpy(buf, "abcdefghijklmnopqrstuvwxyz0123456789");
@@ -407,11 +527,11 @@ static int cmd_stressbuffer(char *args)
                 {
                     PHYSFS_uint32 right = 1 + (PHYSFS_uint32) (35.0 * rand() / (RAND_MAX + 1.0));
                     PHYSFS_uint32 left = 36 - right;
-                    if (PHYSFS_write(f, buf, left, 1) != 1)
+                    if (PHYSFS_writeBytes(f, buf, left) != left)
                     {
-                        printf("PHYSFS_write() failed: %s.\n", PHYSFS_getLastError());
+                        printf("PHYSFS_writeBytes() failed: %s.\n", PHYSFS_getLastError());
                         PHYSFS_close(f);
-                        return(1);
+                        return 1;
                     } /* if */
 
                     rndnum = 1 + (int) (1000.0 * rand() / (RAND_MAX + 1.0));
@@ -421,15 +541,15 @@ static int cmd_stressbuffer(char *args)
                         {
                             printf("PHYSFS_flush() failed: %s.\n", PHYSFS_getLastError());
                             PHYSFS_close(f);
-                            return(1);
+                            return 1;
                         } /* if */
                     } /* if */
 
-                    if (PHYSFS_write(f, buf + left, 1, right) != right)
+                    if (PHYSFS_writeBytes(f, buf + left, right) != right)
                     {
-                        printf("PHYSFS_write() failed: %s.\n", PHYSFS_getLastError());
+                        printf("PHYSFS_writeBytes() failed: %s.\n", PHYSFS_getLastError());
                         PHYSFS_close(f);
-                        return(1);
+                        return 1;
                     } /* if */
 
                     rndnum = 1 + (int) (1000.0 * rand() / (RAND_MAX + 1.0));
@@ -439,7 +559,7 @@ static int cmd_stressbuffer(char *args)
                         {
                             printf("PHYSFS_flush() failed: %s.\n", PHYSFS_getLastError());
                             PHYSFS_close(f);
-                            return(1);
+                            return 1;
                         } /* if */
                     } /* if */
                 } /* for */
@@ -448,7 +568,7 @@ static int cmd_stressbuffer(char *args)
                 {
                     printf("PHYSFS_flush() failed: %s.\n", PHYSFS_getLastError());
                     PHYSFS_close(f);
-                    return(1);
+                    return 1;
                 } /* if */
 
             } /* for */
@@ -456,7 +576,7 @@ static int cmd_stressbuffer(char *args)
             if (!PHYSFS_close(f))
             {
                 printf("PHYSFS_close() failed: %s.\n", PHYSFS_getLastError());
-                return(1);  /* oh well. */
+                return 1;  /* oh well. */
             } /* if */
 
             printf(" ... test file written ...\n");
@@ -464,14 +584,14 @@ static int cmd_stressbuffer(char *args)
             if (f == NULL)
             {
                 printf("Failed to reopen stress file for reading: %s.\n", PHYSFS_getLastError());
-                return(1);
+                return 1;
             } /* if */
 
             if (!PHYSFS_setBuffer(f, num))
             {
                 printf("PHYSFS_setBuffer() failed: %s.\n", PHYSFS_getLastError());
                 PHYSFS_close(f);
-                return(1);
+                return 1;
             } /* if */
 
             for (i = 0; i < 10; i++)
@@ -480,11 +600,11 @@ static int cmd_stressbuffer(char *args)
                 {
                     PHYSFS_uint32 right = 1 + (PHYSFS_uint32) (35.0 * rand() / (RAND_MAX + 1.0));
                     PHYSFS_uint32 left = 36 - right;
-                    if (PHYSFS_read(f, buf2, left, 1) != 1)
+                    if (PHYSFS_readBytes(f, buf2, left) != left)
                     {
-                        printf("PHYSFS_read() failed: %s.\n", PHYSFS_getLastError());
+                        printf("PHYSFS_readBytes() failed: %s.\n", PHYSFS_getLastError());
                         PHYSFS_close(f);
-                        return(1);
+                        return 1;
                     } /* if */
 
                     rndnum = 1 + (int) (1000.0 * rand() / (RAND_MAX + 1.0));
@@ -494,15 +614,15 @@ static int cmd_stressbuffer(char *args)
                         {
                             printf("PHYSFS_flush() failed: %s.\n", PHYSFS_getLastError());
                             PHYSFS_close(f);
-                            return(1);
+                            return 1;
                         } /* if */
                     } /* if */
 
-                    if (PHYSFS_read(f, buf2 + left, 1, right) != right)
+                    if (PHYSFS_readBytes(f, buf2 + left, right) != right)
                     {
-                        printf("PHYSFS_read() failed: %s.\n", PHYSFS_getLastError());
+                        printf("PHYSFS_readBytes() failed: %s.\n", PHYSFS_getLastError());
                         PHYSFS_close(f);
-                        return(1);
+                        return 1;
                     } /* if */
 
                     rndnum = 1 + (int) (1000.0 * rand() / (RAND_MAX + 1.0));
@@ -512,7 +632,7 @@ static int cmd_stressbuffer(char *args)
                         {
                             printf("PHYSFS_flush() failed: %s.\n", PHYSFS_getLastError());
                             PHYSFS_close(f);
-                            return(1);
+                            return 1;
                         } /* if */
                     } /* if */
 
@@ -529,7 +649,7 @@ static int cmd_stressbuffer(char *args)
                             printf("%c", buf2[i]);
                         printf("]\n");
                         PHYSFS_close(f);
-                        return(1);
+                        return 1;
                     } /* if */
                 } /* for */
 
@@ -537,7 +657,7 @@ static int cmd_stressbuffer(char *args)
                 {
                     printf("PHYSFS_flush() failed: %s.\n", PHYSFS_getLastError());
                     PHYSFS_close(f);
-                    return(1);
+                    return 1;
                 } /* if */
 
             } /* for */
@@ -550,7 +670,7 @@ static int cmd_stressbuffer(char *args)
             if (!PHYSFS_close(f))
             {
                 printf("PHYSFS_close() failed: %s.\n", PHYSFS_getLastError());
-                return(1);  /* oh well. */
+                return 1;  /* oh well. */
             } /* if */
 
             PHYSFS_delete("test.txt");
@@ -558,7 +678,7 @@ static int cmd_stressbuffer(char *args)
         } /* else */
     } /* else */
 
-    return(1);
+    return 1;
 } /* cmd_stressbuffer */
 
 
@@ -586,7 +706,7 @@ static int cmd_setsaneconfig(char *args)
     else
         printf("Failure. reason: %s.\n", PHYSFS_getLastError());
 
-    return(1);
+    return 1;
 } /* cmd_setsaneconfig */
 
 
@@ -603,7 +723,7 @@ static int cmd_mkdir(char *args)
     else
         printf("Failure. reason: %s.\n", PHYSFS_getLastError());
 
-    return(1);
+    return 1;
 } /* cmd_mkdir */
 
 
@@ -620,7 +740,7 @@ static int cmd_delete(char *args)
     else
         printf("Failure. reason: %s.\n", PHYSFS_getLastError());
 
-    return(1);
+    return 1;
 } /* cmd_delete */
 
 
@@ -640,7 +760,7 @@ static int cmd_getrealdir(char *args)
     else
         printf("Not found.\n");
 
-    return(1);
+    return 1;
 } /* cmd_getrealdir */
 
 
@@ -656,12 +776,13 @@ static int cmd_exists(char *args)
 
     rc = PHYSFS_exists(args);
     printf("File %sexists.\n", rc ? "" : "does not ");
-    return(1);
+    return 1;
 } /* cmd_exists */
 
 
 static int cmd_isdir(char *args)
 {
+    PHYSFS_Stat statbuf;
     int rc;
 
     if (*args == '\"')
@@ -670,14 +791,17 @@ static int cmd_isdir(char *args)
         args[strlen(args) - 1] = '\0';
     } /* if */
 
-    rc = PHYSFS_isDirectory(args);
+    rc = PHYSFS_stat(args, &statbuf);
+    if (rc)
+        rc = (statbuf.filetype == PHYSFS_FILETYPE_DIRECTORY);
     printf("File %s a directory.\n", rc ? "is" : "is NOT");
-    return(1);
+    return 1;
 } /* cmd_isdir */
 
 
 static int cmd_issymlink(char *args)
 {
+    PHYSFS_Stat statbuf;
     int rc;
 
     if (*args == '\"')
@@ -686,9 +810,11 @@ static int cmd_issymlink(char *args)
         args[strlen(args) - 1] = '\0';
     } /* if */
 
-    rc = PHYSFS_isSymbolicLink(args);
+    rc = PHYSFS_stat(args, &statbuf);
+    if (rc)
+        rc = (statbuf.filetype == PHYSFS_FILETYPE_SYMLINK);
     printf("File %s a symlink.\n", rc ? "is" : "is NOT");
-    return(1);
+    return 1;
 } /* cmd_issymlink */
 
 
@@ -714,7 +840,7 @@ static int cmd_cat(char *args)
                 printf("failed to set file buffer. Reason: [%s].\n",
                         PHYSFS_getLastError());
                 PHYSFS_close(f);
-                return(1);
+                return 1;
             } /* if */
         } /* if */
 
@@ -723,7 +849,7 @@ static int cmd_cat(char *args)
             char buffer[128];
             PHYSFS_sint64 rc;
             PHYSFS_sint64 i;
-            rc = PHYSFS_read(f, buffer, 1, sizeof (buffer));
+            rc = PHYSFS_readBytes(f, buffer, sizeof (buffer));
 
             for (i = 0; i < rc; i++)
                 fputc((int) buffer[i], stdout);
@@ -737,14 +863,212 @@ static int cmd_cat(char *args)
                            PHYSFS_getLastError());
                 } /* if */
                 PHYSFS_close(f);
-                return(1);
+                return 1;
             } /* if */
         } /* while */
     } /* else */
 
-    return(1);
+    return 1;
 } /* cmd_cat */
 
+static int cmd_cat2(char *args)
+{
+    PHYSFS_File *f1 = NULL;
+    PHYSFS_File *f2 = NULL;
+    char *fname1;
+    char *fname2;
+    char *ptr;
+
+    fname1 = args;
+    if (*fname1 == '\"')
+    {
+        fname1++;
+        ptr = strchr(fname1, '\"');
+        if (ptr == NULL)
+        {
+            printf("missing string terminator in argument.\n");
+            return 1;
+        } /* if */
+        *(ptr) = '\0';
+    } /* if */
+    else
+    {
+        ptr = strchr(fname1, ' ');
+        *ptr = '\0';
+    } /* else */
+
+    fname2 = ptr + 1;
+    if (*fname2 == '\"')
+    {
+        fname2++;
+        ptr = strchr(fname2, '\"');
+        if (ptr == NULL)
+        {
+            printf("missing string terminator in argument.\n");
+            return 1;
+        } /* if */
+        *(ptr) = '\0';
+    } /* if */
+
+    if ((f1 = PHYSFS_openRead(fname1)) == NULL)
+        printf("failed to open '%s'. Reason: [%s].\n", fname1, PHYSFS_getLastError());
+    else if ((f2 = PHYSFS_openRead(fname2)) == NULL)
+        printf("failed to open '%s'. Reason: [%s].\n", fname2, PHYSFS_getLastError());
+    else
+    {
+        char *buffer1 = NULL;
+        size_t buffer1len = 0;
+        char *buffer2 = NULL;
+        size_t buffer2len = 0;
+        char *ptr = NULL;
+        size_t i;
+
+        if (do_buffer_size)
+        {
+            if (!PHYSFS_setBuffer(f1, do_buffer_size))
+            {
+                printf("failed to set file buffer for '%s'. Reason: [%s].\n",
+                        fname1, PHYSFS_getLastError());
+                PHYSFS_close(f1);
+                PHYSFS_close(f2);
+                return 1;
+            } /* if */
+            else if (!PHYSFS_setBuffer(f2, do_buffer_size))
+            {
+                printf("failed to set file buffer for '%s'. Reason: [%s].\n",
+                        fname2, PHYSFS_getLastError());
+                PHYSFS_close(f1);
+                PHYSFS_close(f2);
+                return 1;
+            } /* if */
+        } /* if */
+
+
+        do
+        {
+            int readlen = 128;
+            PHYSFS_sint64 rc;
+
+            ptr = realloc(buffer1, buffer1len + readlen);
+            if (!ptr)
+            {
+                printf("(Out of memory.)\n\n");
+                free(buffer1);
+                free(buffer2);
+                PHYSFS_close(f1);
+                PHYSFS_close(f2);
+                return 1;
+            } /* if */
+
+            buffer1 = ptr;
+            rc = PHYSFS_readBytes(f1, buffer1 + buffer1len, readlen);
+            if (rc < 0)
+            {
+                printf("(Error condition in reading '%s'. Reason: [%s])\n\n",
+                       fname1, PHYSFS_getLastError());
+                free(buffer1);
+                free(buffer2);
+                PHYSFS_close(f1);
+                PHYSFS_close(f2);
+                return 1;
+            } /* if */
+            buffer1len += (size_t) rc;
+
+            ptr = realloc(buffer2, buffer2len + readlen);
+            if (!ptr)
+            {
+                printf("(Out of memory.)\n\n");
+                free(buffer1);
+                free(buffer2);
+                PHYSFS_close(f1);
+                PHYSFS_close(f2);
+                return 1;
+            } /* if */
+
+            buffer2 = ptr;
+            rc = PHYSFS_readBytes(f2, buffer2 + buffer2len, readlen);
+            if (rc < 0)
+            {
+                printf("(Error condition in reading '%s'. Reason: [%s])\n\n",
+                       fname2, PHYSFS_getLastError());
+                free(buffer1);
+                free(buffer2);
+                PHYSFS_close(f1);
+                PHYSFS_close(f2);
+                return 1;
+            } /* if */
+            buffer2len += (size_t) rc;
+        } while (!PHYSFS_eof(f1) || !PHYSFS_eof(f2));
+
+        printf("file '%s' ...\n\n", fname1);
+        for (i = 0; i < buffer1len; i++)
+            fputc((int) buffer1[i], stdout);
+        free(buffer1);
+
+        printf("\n\nfile '%s' ...\n\n", fname2);
+        for (i = 0; i < buffer2len; i++)
+            fputc((int) buffer2[i], stdout);
+        free(buffer2);
+
+        printf("\n\n");
+    } /* else */
+
+    if (f1)
+        PHYSFS_close(f1);
+
+    if (f2)
+        PHYSFS_close(f2);
+
+    return 1;
+} /* cmd_cat2 */
+
+
+#define CRC32_BUFFERSIZE 512
+static int cmd_crc32(char *args)
+{
+    PHYSFS_File *f;
+
+    if (*args == '\"')
+    {
+        args++;
+        args[strlen(args) - 1] = '\0';
+    } /* if */
+
+    f = PHYSFS_openRead(args);
+    if (f == NULL)
+        printf("failed to open. Reason: [%s].\n", PHYSFS_getLastError());
+    else
+    {
+        PHYSFS_uint8 buffer[CRC32_BUFFERSIZE];
+        PHYSFS_uint32 crc = -1;
+        PHYSFS_sint64 bytesread;
+
+        while ((bytesread = PHYSFS_readBytes(f, buffer, CRC32_BUFFERSIZE)) > 0)
+        {
+            PHYSFS_uint32 i, bit;
+            for (i = 0; i < bytesread; i++)
+            {
+                for (bit = 0; bit < 8; bit++, buffer[i] >>= 1)
+                    crc = (crc >> 1) ^ (((crc ^ buffer[i]) & 1) ? 0xEDB88320 : 0);
+            } /* for */
+        } /* while */
+
+        if (bytesread < 0)
+        {
+            printf("error while reading. Reason: [%s].\n",
+                   PHYSFS_getLastError());
+            return 1;
+        } /* if */
+
+        PHYSFS_close(f);
+
+        crc ^= -1;
+        printf("CRC32 for %s: 0x%08X\n", args, crc);
+    } /* else */
+
+    return 1;
+} /* cmd_crc32 */
+
 
 static int cmd_filelength(char *args)
 {
@@ -770,10 +1094,9 @@ static int cmd_filelength(char *args)
         PHYSFS_close(f);
     } /* else */
 
-    return(1);
+    return 1;
 } /* cmd_filelength */
 
-
 #define WRITESTR "The cat sat on the mat.\n\n"
 
 static int cmd_append(char *args)
@@ -801,12 +1124,12 @@ static int cmd_append(char *args)
                 printf("failed to set file buffer. Reason: [%s].\n",
                         PHYSFS_getLastError());
                 PHYSFS_close(f);
-                return(1);
+                return 1;
             } /* if */
         } /* if */
 
         bw = strlen(WRITESTR);
-        rc = PHYSFS_write(f, WRITESTR, 1, bw);
+        rc = PHYSFS_writeBytes(f, WRITESTR, bw);
         if (rc != bw)
         {
             printf("Wrote (%d) of (%d) bytes. Reason: [%s].\n",
@@ -820,7 +1143,7 @@ static int cmd_append(char *args)
         PHYSFS_close(f);
     } /* else */
 
-    return(1);
+    return 1;
 } /* cmd_append */
 
 
@@ -849,12 +1172,12 @@ static int cmd_write(char *args)
                 printf("failed to set file buffer. Reason: [%s].\n",
                         PHYSFS_getLastError());
                 PHYSFS_close(f);
-                return(1);
+                return 1;
             } /* if */
         } /* if */
 
         bw = strlen(WRITESTR);
-        rc = PHYSFS_write(f, WRITESTR, 1, bw);
+        rc = PHYSFS_writeBytes(f, WRITESTR, bw);
         if (rc != bw)
         {
             printf("Wrote (%d) of (%d) bytes. Reason: [%s].\n",
@@ -868,34 +1191,79 @@ static int cmd_write(char *args)
         PHYSFS_close(f);
     } /* else */
 
-    return(1);
+    return 1;
 } /* cmd_write */
 
 
-static void modTimeToStr(PHYSFS_sint64 modtime, char *modstr, size_t strsize)
+static char* modTimeToStr(PHYSFS_sint64 modtime, char *modstr, size_t strsize)
 {
-    time_t t = (time_t) modtime;
-    char *str = ctime(&t);
-    strncpy(modstr, str, strsize);
+    if (modtime < 0)
+        strncpy(modstr, "Unknown\n", strsize);
+    else
+    {
+        time_t t = (time_t) modtime;
+        char *str = ctime(&t);
+        strncpy(modstr, str, strsize);
+    } /* else */
+
     modstr[strsize-1] = '\0';
+    return modstr;
 } /* modTimeToStr */
 
 
 static int cmd_getlastmodtime(char *args)
 {
-    PHYSFS_sint64 rc = PHYSFS_getLastModTime(args);
-    if (rc == -1)
+    PHYSFS_Stat statbuf;
+    if (!PHYSFS_stat(args, &statbuf))
         printf("Failed to determine. Reason: [%s].\n", PHYSFS_getLastError());
     else
     {
         char modstr[64];
-        modTimeToStr(rc, modstr, sizeof (modstr));
-        printf("Last modified: %s (%ld).\n", modstr, (long) rc);
+        modTimeToStr(statbuf.modtime, modstr, sizeof (modstr));
+        printf("Last modified: %s (%ld).\n", modstr, (long) statbuf.modtime);
     } /* else */
 
-    return(1);
+    return 1;
 } /* cmd_getLastModTime */
 
+static int cmd_stat(char *args)
+{
+    PHYSFS_Stat stat;
+    char timestring[65];
+
+    if (*args == '\"')
+    {
+        args++;
+        args[strlen(args) - 1] = '\0';
+    } /* if */
+
+    if(!PHYSFS_stat(args, &stat))
+    {
+        printf("failed to stat. Reason [%s].\n", PHYSFS_getLastError());
+        return 1;
+    } /* if */
+
+    printf("Filename: %s\n", args);
+    printf("Size %d\n",(int) stat.filesize);
+
+    if(stat.filetype == PHYSFS_FILETYPE_REGULAR)
+        printf("Type: File\n");
+    else if(stat.filetype == PHYSFS_FILETYPE_DIRECTORY)
+        printf("Type: Directory\n");
+    else if(stat.filetype == PHYSFS_FILETYPE_SYMLINK)
+        printf("Type: Symlink\n");
+    else
+        printf("Type: Unknown\n");
+
+    printf("Created at: %s", modTimeToStr(stat.createtime, timestring, 64));
+    printf("Last modified at: %s", modTimeToStr(stat.modtime, timestring, 64));
+    printf("Last accessed at: %s", modTimeToStr(stat.accesstime, timestring, 64));
+    printf("Readonly: %s\n", stat.readonly ? "true" : "false");
+
+    return 1;
+} /* cmd_filelength */
+
+
 
 /* must have spaces trimmed prior to this call. */
 static int count_args(const char *str)
@@ -915,7 +1283,7 @@ static int count_args(const char *str)
         retval++;
     } /* if */
 
-    return(retval);
+    return retval;
 } /* count_args */
 
 
@@ -938,7 +1306,10 @@ static const command_info commands[] =
     { "deinit",         cmd_deinit,         0, NULL                         },
     { "addarchive",     cmd_addarchive,     2, "<archiveLocation> <append>" },
     { "mount",          cmd_mount,          3, "<archiveLocation> <mntpoint> <append>" },
+    { "mountmem",       cmd_mount_mem,      3, "<archiveLocation> <mntpoint> <append>" },
+    { "mounthandle",    cmd_mount_handle,   3, "<archiveLocation> <mntpoint> <append>" },
     { "removearchive",  cmd_removearchive,  1, "<archiveLocation>"          },
+    { "unmount",        cmd_removearchive,  1, "<archiveLocation>"          },
     { "enumerate",      cmd_enumerate,      1, "<dirToEnumerate>"           },
     { "ls",             cmd_enumerate,      1, "<dirToEnumerate>"           },
     { "getlasterror",   cmd_getlasterror,   0, NULL                         },
@@ -947,6 +1318,7 @@ static const command_info commands[] =
     { "getsearchpath",  cmd_getsearchpath,  0, NULL                         },
     { "getbasedir",     cmd_getbasedir,     0, NULL                         },
     { "getuserdir",     cmd_getuserdir,     0, NULL                         },
+    { "getprefdir",     cmd_getprefdir,     2, "<org> <app>"                },
     { "getwritedir",    cmd_getwritedir,    0, NULL                         },
     { "setwritedir",    cmd_setwritedir,    1, "<newWriteDir>"              },
     { "permitsymlinks", cmd_permitsyms,     1, "<1or0>"                     },
@@ -958,12 +1330,16 @@ static const command_info commands[] =
     { "isdir",          cmd_isdir,          1, "<fileToCheck>"              },
     { "issymlink",      cmd_issymlink,      1, "<fileToCheck>"              },
     { "cat",            cmd_cat,            1, "<fileToCat>"                },
+    { "cat2",           cmd_cat2,           2, "<fileToCat1> <fileToCat2>"  },
     { "filelength",     cmd_filelength,     1, "<fileToCheck>"              },
+    { "stat",           cmd_stat,           1, "<fileToStat>"               },
     { "append",         cmd_append,         1, "<fileToAppend>"             },
     { "write",          cmd_write,          1, "<fileToCreateOrTrash>"      },
     { "getlastmodtime", cmd_getlastmodtime, 1, "<fileToExamine>"            },
     { "setbuffer",      cmd_setbuffer,      1, "<bufferSize>"               },
     { "stressbuffer",   cmd_stressbuffer,   1, "<bufferSize>"               },
+    { "crc32",          cmd_crc32,          1, "<fileToHash>"               },
+    { "getmountpoint",  cmd_getmountpoint,  1, "<dir>"                      },
     { NULL,             NULL,              -1, NULL                         }
 };
 
@@ -985,7 +1361,7 @@ static int cmd_help(char *args)
     for (i = commands; i->cmd != NULL; i++)
         output_usage("  -", i);
 
-    return(1);
+    return 1;
 } /* output_cmd_help */
 
 
@@ -1037,14 +1413,14 @@ static int process_command(char *complete_cmd)
     if (complete_cmd == NULL)  /* can happen if user hits CTRL-D, etc. */
     {
         printf("\n");
-        return(0);
+        return 0;
     } /* if */
 
     cmd_copy = (char *) malloc(strlen(complete_cmd) + 1);
     if (cmd_copy == NULL)
     {
         printf("\n\n\nOUT OF MEMORY!\n\n\n");
-        return(0);
+        return 0;
     } /* if */
 
     trim_command(complete_cmd, cmd_copy);
@@ -1084,7 +1460,7 @@ static int process_command(char *complete_cmd)
     } /* if */
 
     free(cmd_copy);
-    return(rc);
+    return rc;
 } /* process_command */
 
 
@@ -1157,7 +1533,7 @@ int main(int argc, char **argv)
     if (!PHYSFS_init(argv[0]))
     {
         printf("PHYSFS_init() failed!\n  reason: %s.\n", PHYSFS_getLastError());
-        return(1);
+        return 1;
     } /* if */
 
     output_versions();
@@ -1166,6 +1542,7 @@ int main(int argc, char **argv)
     open_history_file();
 
     printf("Enter commands. Enter \"help\" for instructions.\n");
+    fflush(stdout);
 
     do
     {
@@ -1176,6 +1553,7 @@ int main(int argc, char **argv)
         buf = (char *) malloc(512);
         memset(buf, '\0', 512);
         printf("> ");
+        fflush(stdout);
         for (i = 0; i < 511; i++)
         {
             int ch = fgetc(stdin);
@@ -1202,6 +1580,7 @@ int main(int argc, char **argv)
 #endif
 
         rc = process_command(buf);
+        fflush(stdout);
         if (buf != NULL)
             free(buf);
     } while (rc);
@@ -1217,7 +1596,7 @@ int main(int argc, char **argv)
     printf(" it makes you shoot teh railgun bettar.\n");
 */
 
-    return(0);
+    return 0;
 } /* main */
 
 /* end of test_physfs.c ... */
